/* xeyes+  Version 2.01 */
/*
 * Copyleft 1994 Software Research Academy Japan.
 *
 *                        Presented by    Yusuke Yoneyama
 *                        Special thanks  Tetsu  Satoh   
 *
 */

/*
 *   follows the mouse cursor around
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include <sys/time.h>
#include <X11/Xatom.h>
#ifdef XPM
#include <X11/xpm.h>
#else
#include "xpm.h"
#endif /* XPM */

#include "eyebase.xpm"
#include "eyebasemask.xbm"
#ifndef V1
#include "eye.xpm"
#include "eyemask.xbm"
#else /* !V1 */
#include "eyeV1.xpm"
#include "eyemaskV1.xbm"
#endif /* V1 */

#define EYE_WIDTH      eyemask_width
#define EYE_HEIGHT     eyemask_height
#define MAX_EYES_X     32
#define MAX_EYES_Y     32
#define EYEBASE_WIDTH  160
#define EYEBASE_HEIGHT 160
#define CENTER_X       82
#define CENTER_Y       73
#define EYE_MOVER_X    24
#define EYE_MOVER_Y    13
#define DRAW_X         20
#define DRAW_Y         35
#define DRAW_WIDTH     120
#define DRAW_HEIGHT    75
#define MISS_TIME      200

int rooteyes =  False; /* root window mode flag        */
int wmeyes   =  False; /* window manager mode flag     */
int speed    =  20;    /* default speed (20/100 usec.) */
int smooth   =  False; /* moving smooth flag           */

struct xy {
      int x;
      int y;
};

void xeyesp(
#if NeedFunctionPrototypes
    char *
#endif /* NeedFunctionPrototypes */
);

void DrawEye(
#if NeedFunctionPrototypes
    Display *,
    Window,
    GC,
    int,
    int, 
    int,
    int, 
    int, 
    int
#endif /* NeedFunctionPrototypes */
);

#ifdef USLEEP
#define u_sleep(x) usleep((x)*10.0)
#else /* !USLEEP */
 void u_sleep(
#if NeedFunctionPrototypes
   double
#endif /* NeedFunctionPrototypes */
 );
#endif /* USLEEP */

void usage(
#if NeedFunctionPrototypes
  void
#endif /* NeedFunctionPrototypes */
);

void geteyepos(
#if NeedFunctionPrototypes
  int *, int *, int, int, int, int
#endif /* NeedFunctionPrototypes */
);



void usage()
{
  printf("Usage: xeyes+");
  printf(" [-root] [-wm] [-speed n] [-smooth] [-display <display>]\n");
  exit(0);
}


void main(argc, argv)
int argc;
char **argv;
{
  int i;
  char *display_name = NULL;

  if(argc > 1) {

    /*** check options ***/

    for(i=1;i<argc;i++) {
      if(strcmp(argv[i], "-display")==0) {
        i++;
        if(i >= argc) usage();
        display_name = argv[i];
      }
      else if(strcmp(argv[i], "-root")==0) {
        rooteyes = True;
      }
      else if(strcmp(argv[i], "-wm")==0) {
        wmeyes = True;
      }
      else if(strcmp(argv[i], "-speed")==0) {
	i++;
	if(i >= argc) usage();
        speed = atoi(argv[i]);
        if(speed < 0 || speed > 100) usage();
      }
      else if(strcmp(argv[i], "-smooth")==0) {
        smooth = True;
      }
      else {
        usage();
      }
    }
  }

  xeyesp(display_name);

  exit(0);
}


Display *dpy;
Window root;
int screen;
int display_width, display_height;
int default_depth;
Pixmap eyebase_xpm, eyebasemask_xpm, eye_xpm, eyemask_xpm;
Pixmap tmpxpm1;
#ifndef V1
 Pixmap tmpxpm2;
#endif /* !v1 */

void xeyesp(display_name)
char *display_name;
{
  XWindowAttributes root_attr;
  XpmAttributes xpm_attr;
  XSetWindowAttributes xset_attr;
  GC gc;
  Window eyes;
  Window dw,cw;
  XEvent e;
  XSizeHints shint;
  XWindowChanges wc;
  int eyesx, eyesy;
  struct xy xeyesxy[MAX_EYES_Y][MAX_EYES_X];
  int i, j;
  int x, y, rx, ry, wx, wy;
  int dx, dy, dl;
  int ox, oy;
  int draweyes;
  unsigned int mask;
  int ef = 0;
  

  dpy = XOpenDisplay(display_name);
  if (dpy == NULL) {
    fprintf(stderr, "xeyes+:  unable to open display '%s'\n",
            XDisplayName(display_name));
    return;
  }
  
  screen = DefaultScreen(dpy);
  root = RootWindow(dpy, screen);

  display_width  = DisplayWidth(dpy, screen);
  display_height = DisplayHeight(dpy, screen);
  default_depth = DefaultDepth(dpy, screen);

  srand((int)time((time_t *)NULL));
  XGetWindowAttributes(dpy, root, &root_attr);
  xpm_attr.colormap = root_attr.colormap;
  xpm_attr.valuemask = XpmSize | XpmReturnPixels | XpmColormap;
  XpmCreatePixmapFromData(dpy, root, (char **)&eyebase,
                          &eyebase_xpm, NULL, &xpm_attr);
  XpmCreatePixmapFromData(dpy,root,(char **)&eye,
                          &eye_xpm, NULL, &xpm_attr);
  eyemask_xpm = XCreatePixmapFromBitmapData(dpy, root, (char *)eyemask_bits,
                          eyemask_width, eyemask_height,
                          AllPlanes, 0, default_depth);
  eyebasemask_xpm = XCreatePixmapFromBitmapData(dpy, root, 
                          (char *)eyebasemask_bits, eyebasemask_width,
                          eyebasemask_height, AllPlanes, 0, default_depth);
  tmpxpm1 = XCreatePixmap(dpy, root, DRAW_WIDTH, DRAW_HEIGHT,
                          default_depth);
#ifndef V1
  tmpxpm2 = XCreatePixmap(dpy, root, DRAW_WIDTH, DRAW_HEIGHT,
                          default_depth);
#endif /* !V1 */

  eyesx = display_width / EYEBASE_WIDTH + 
          ((display_width % EYEBASE_WIDTH) ? 1 : 0);
  eyesy = display_height / EYEBASE_HEIGHT +
          ((display_height % EYEBASE_HEIGHT) ? 1 : 0);
  
  gc = XCreateGC(dpy, root, 0, 0);
  
  if(rooteyes == False) {
    eyes = XCreateSimpleWindow(dpy, root, 0, 0, 
                               display_width, display_height,0, 1, 0);
    XSelectInput(dpy, eyes, ExposureMask | ResizeRedirectMask);
    if(wmeyes == False) {
        xset_attr.override_redirect = True;
        XChangeWindowAttributes(dpy, eyes, CWOverrideRedirect, &xset_attr);
        XMapWindow(dpy, eyes);
        XLowerWindow(dpy, eyes);
    }
    else {
        shint.max_width  = display_width;
        shint.max_height = display_height;
        shint.flags = PMaxSize;
        XSetNormalHints(dpy, eyes, &shint);
        XStoreName(dpy, eyes, "xeyes+");
        xset_attr.override_redirect = True;
        XChangeWindowAttributes(dpy, eyes, CWOverrideRedirect, &xset_attr);
        XResizeWindow(dpy, eyes, EYEBASE_WIDTH * 2, EYEBASE_HEIGHT);
        wc.width = EYEBASE_WIDTH * 2;
        wc.height = EYEBASE_HEIGHT;
        XConfigureWindow(dpy, eyes, CWWidth | CWHeight, &wc);
        xset_attr.override_redirect = False;
        XChangeWindowAttributes(dpy, eyes, CWOverrideRedirect, &xset_attr);
        XMapWindow(dpy, eyes);
        eyesx = 2;
        eyesy = 1;
    }
  }
  else {
      eyes = root;
  }

  XSetWindowBackgroundPixmap(dpy, eyes, eyebase_xpm);
  XClearWindow(dpy, eyes);

  for(j = 0; j < eyesy; j++) {
    for(i = 0; i < eyesx; i++) {
      DrawEye(dpy, eyes, gc, i, j, 0, 0, 0, 0);
      xeyesxy[j][i].x = 0;
      xeyesxy[j][i].y = 0;
    }
  }

  XFlush(dpy);

  /*** main loop ***/
  while(1) {
    
    u_sleep((double)speed / 100.0);

    if(XCheckTypedWindowEvent(dpy, eyes, ResizeRequest, &e)) {

      /*** Change Window Configuration ***/

      while(XCheckTypedWindowEvent(dpy, eyes, ResizeRequest, &e));
      xset_attr.override_redirect = True;
      XChangeWindowAttributes(dpy, eyes, CWOverrideRedirect, &xset_attr);
      XResizeWindow(dpy, eyes, e.xresizerequest.width,
                    e.xresizerequest.height);
      wc.width = e.xresizerequest.width;
      wc.height = e.xresizerequest.height;
      XConfigureWindow(dpy, eyes, CWWidth | CWHeight, &wc);
      xset_attr.override_redirect = False;
      XChangeWindowAttributes(dpy, eyes, CWOverrideRedirect, &xset_attr);
      ox = eyesx;
      oy = eyesy;
      eyesx = wc.width / EYEBASE_WIDTH + 
              ((wc.width % EYEBASE_WIDTH) ? 1 : 0);
      eyesy = wc.height / EYEBASE_HEIGHT +
              ((wc.height % EYEBASE_HEIGHT) ? 1 : 0);
      /* initialize new area */
      for(j = 0; j < eyesy; j++) {
        for(i = 0; i < eyesx; i++) {
          if(i >= ox || j >= oy) {
            xeyesxy[j][i].x = 0;
            xeyesxy[j][i].y = 0;
          }
        }
      }
    }

    if(XCheckTypedWindowEvent(dpy, eyes, Expose, &e)) {

      /*** redraw all eyes ***/

      while(XCheckTypedWindowEvent(dpy, eyes, Expose, &e));
      if(rooteyes == False && wmeyes == False) {
      	XLowerWindow(dpy,eyes);
      }
      for(j = 0; j < eyesy; j++) {
        for(i = 0; i < eyesx; i++) {
          DrawEye(dpy, eyes, gc, i, j, xeyesxy[j][i].x, xeyesxy[j][i].y,
                  xeyesxy[j][i].x, xeyesxy[j][i].y);
        }
      }
    }
    else {
      while(XPending(dpy)) {
        XNextEvent(dpy, &e);
      }
    }

    XQueryPointer(dpy, root, &dw, &cw, &rx, &ry, &wx, &wy, &mask);
    
    if(cw == eyes || (rooteyes == 1 && cw == 0) || (wmeyes == 1)) { 
      /*** on the eyes window ***/

      if(ef == 0) {
      	if(rooteyes == False && wmeyes == False) {
          XLowerWindow(dpy, eyes);
        }
        ef = 1;
      }

      if(wmeyes == True) {
        XQueryPointer(dpy, eyes, &dw, &cw, &rx, &ry, &wx, &wy, &mask);
        x = wx / EYEBASE_WIDTH;
        y = wy / EYEBASE_HEIGHT;
      }
      else {
        x = rx / EYEBASE_WIDTH;
        y = ry / EYEBASE_HEIGHT;
      }
      if(eyesx * eyesy > 10) {
      	draweyes = eyesx * eyesy / 2;
      }
      else {
        draweyes = eyesx * eyesy + 1;
      }
      for(i = 0; i < draweyes; i++) {
        if(x >= eyesx || y >= eyesy) {
          x = rand() % eyesx;
	  y = rand() % eyesy;
          continue;
        }
      	if(wmeyes == True) {
          geteyepos(&dx, &dy, x * EYEBASE_WIDTH + CENTER_X ,
                    y * EYEBASE_HEIGHT + CENTER_Y , wx, wy);
        }
        else {
          geteyepos(&dx, &dy, x * EYEBASE_WIDTH + CENTER_X,
                    y * EYEBASE_HEIGHT + CENTER_Y, rx, ry);
        }
        if(dx == xeyesxy[y][x].x && dy == xeyesxy[y][x].y) {
          x = rand() % eyesx;
	  y = rand() % eyesy;
          continue;
        }
        ox = xeyesxy[y][x].x;
        oy = xeyesxy[y][x].y;
        if(smooth == True) {
          dl = dx - xeyesxy[y][x].x;
          if(dl > 0)
            xeyesxy[y][x].x += (int)sqrt((double)abs(dl));
          else
            xeyesxy[y][x].x -= (int)sqrt((double)abs(dl));
          dl = dy - xeyesxy[y][x].y;
          if(dl > 0)
            xeyesxy[y][x].y += (int)sqrt((double)abs(dl));
          else
            xeyesxy[y][x].y -= (int)sqrt((double)abs(dl));
        }
        else {
          xeyesxy[y][x].x = dx;
          xeyesxy[y][x].y = dy;
        }
        DrawEye(dpy, eyes, gc, x, y, ox, oy,
                xeyesxy[y][x].x, xeyesxy[y][x].y);
	x = rand() % eyesx;
	y = rand() % eyesy;
      }
    }
    else { /*** on the other window ***/
      ef = 0;
      for(j = 0; j < eyesy; j++) {
	for(i = 0; i < eyesx; i++) {
	  if(rand() % 30 < 29) continue;
          x = rand() % eyesx;
          y = rand() % eyesy;
          geteyepos(&dx, &dy, x * EYEBASE_WIDTH + CENTER_X,
                    y * EYEBASE_HEIGHT + CENTER_Y, rand() % display_width,
                    rand() % display_height);
          if(xeyesxy[y][x].x == dx && xeyesxy[y][x].y == dy) {
            continue;
          }
          ox = xeyesxy[y][x].x;
          oy = xeyesxy[y][x].y;
          if(smooth == True) {
            dl = dx - xeyesxy[y][x].x;
            if(dl > 0)
              xeyesxy[y][x].x += (int)sqrt((double)abs(dl));
            else
              xeyesxy[y][x].x -= (int)sqrt((double)abs(dl));
            dl = dy - xeyesxy[y][x].y;
            if(dl > 0)
              xeyesxy[y][x].y += (int)sqrt((double)abs(dl));
            else
              xeyesxy[y][x].y -= (int)sqrt((double)abs(dl));
          }
          else {
            xeyesxy[y][x].x = dx;
            xeyesxy[y][x].y = dy;
          }
          DrawEye(dpy, eyes, gc, x, y, ox, oy,
                  xeyesxy[y][x].x, xeyesxy[y][x].y);
	}
      }
    }

    XFlush(dpy);
  }
}


void DrawEye(dpy, win, gc, eyex, eyey, oldx, oldy, newx, newy)
Display *dpy;
Window  win;
GC      gc;
int     eyex, eyey, oldx, oldy, newx, newy;
{
  XSetFunction(dpy, gc, GXcopy);
  XCopyArea(dpy, eyebase_xpm, tmpxpm1, gc, DRAW_X, DRAW_Y,
            DRAW_WIDTH, DRAW_HEIGHT, 0, 0);
#ifndef V1
  XCopyArea(dpy, eyebase_xpm, tmpxpm2, gc, DRAW_X, DRAW_Y,
            DRAW_WIDTH, DRAW_HEIGHT, 0, 0);
#endif /* V1 */

  XSetFunction(dpy, gc, GXand);
  XCopyArea(dpy, eyemask_xpm, tmpxpm1, gc, 0, 0, EYE_WIDTH, EYE_HEIGHT,
            CENTER_X - EYE_WIDTH  / 2 - DRAW_X + newx,
            CENTER_Y - EYE_HEIGHT / 2 - DRAW_Y + newy);
  XSetFunction(dpy, gc, GXor);
  XCopyArea(dpy, eye_xpm, tmpxpm1, gc, 0, 0, EYE_WIDTH, EYE_HEIGHT,
            CENTER_X - EYE_WIDTH  / 2 - DRAW_X + newx,
            CENTER_Y - EYE_HEIGHT / 2 - DRAW_Y + newy);

#ifndef V1
  XSetFunction(dpy, gc, GXand);
  XCopyArea(dpy, eyebasemask_xpm, tmpxpm1, gc, DRAW_X, DRAW_Y,
            DRAW_WIDTH, DRAW_HEIGHT, 0, 0);
  XSetFunction(dpy, gc, GXandInverted);
  XCopyArea(dpy, eyebasemask_xpm, tmpxpm2, gc, DRAW_X, DRAW_Y,
            DRAW_WIDTH, DRAW_HEIGHT, 0, 0);
  XSetFunction(dpy, gc, GXor);
  XCopyArea(dpy, tmpxpm1, tmpxpm2, gc, 0, 0, DRAW_WIDTH, DRAW_HEIGHT, 0, 0);

  XSetFunction(dpy, gc, GXcopy);
  XCopyArea(dpy, tmpxpm2, win, gc, 0, 0, DRAW_WIDTH, DRAW_HEIGHT,
            eyex * EYEBASE_WIDTH + DRAW_X, eyey * EYEBASE_HEIGHT + DRAW_Y);
#else /* V1 */
  XSetFunction(dpy, gc, GXcopy);
  XCopyArea(dpy, tmpxpm1, win, gc, 0, 0, DRAW_WIDTH, DRAW_HEIGHT,
            eyex * EYEBASE_WIDTH + DRAW_X, eyey * EYEBASE_HEIGHT + DRAW_Y);
#endif /* !V1 */

  return;
}


void geteyepos(eyex, eyey, rootx, rooty, mousex, mousey)
int *eyex, *eyey;
int rootx, rooty;
int mousex, mousey;
{
  int rx, ry;
  double th, x, y, d;
  
  rx = mousex - rootx;
  ry = mousey - rooty;
  if(mousex <= rootx + EYE_MOVER_X && mousex >= rootx - EYE_MOVER_X &&
     mousey <= rooty + EYE_MOVER_Y && mousey >= rooty - EYE_MOVER_Y ) {
      if(sqrt(rx * rx + ry * ry) <= EYE_MOVER_X) {
        *eyex = mousex - rootx;
        *eyey = mousey - rooty;
        return;
      }
  }
  if(rx == 0) {
    rx = 1;
  }
  th = fabs(atan((double)ry / (double)rx));
#ifndef V1
  if(abs(rx) > display_width / 3)
    if(abs(rx) > display_width * 2 / 3)
      d = 1.8;
    else
      d = 1.4;
  else
#endif
    d = 1.0;
  x = (int)((double)EYE_MOVER_X* d * cos(th));

#ifndef V1
  if(abs(ry) > display_height / 3)
    if(abs(ry) > display_height * 2 / 3)
      d = 1.8;
    else
      d = 1.4;
  else
#endif
    d = 1.0;
  y = (int)((double)EYE_MOVER_Y* d * sin(th));

  if(rx < 0 )
        *eyex = (int)(-x);
  else  *eyex = (int)x;
  if(ry < 0 ) 
        *eyey = (int)(-y);
  else  *eyey = (int)y;

  if((rand() % MISS_TIME) == 0) {
  	*eyex = - *eyex;
  	*eyey = - *eyey;
  }

  return;
}


#ifndef USLEEP
void u_sleep(sec)
     double  sec;
{
  int fds = 0;
  struct timeval timeout;
  
  timeout.tv_sec = sec;
  timeout.tv_usec = (sec - timeout.tv_sec) * 1000000.0;
  select(0, (fd_set *)&fds, (fd_set *)&fds, (fd_set *)&fds, &timeout);

  return;
}
#endif /* !USLEEP */
